import asyncio
import json
from pathlib import Path

from trame_client.widgets.core import TrameDefault

from pan3d import catalogs as pan3d_catalogs
from pan3d.ui.catalog_search import CatalogSearch
from pan3d.xarray.algorithm import vtkXArrayRectilinearSource
from trame.app import get_server
from trame.decorators import TrameApp, change
from trame.ui.vuetify3 import SinglePageWithDrawerLayout
from trame.widgets import html
from trame.widgets import vuetify3 as v3


@TrameApp()
class CatalogBrowser:
    def __init__(self, server=None):
        self.server = get_server(server)
        self.current_event_loop = asyncio.get_event_loop()

        # dev setup
        if self.server.hot_reload:
            self.ctrl.on_server_reload.add(self._build_ui)

        # Xarray helper
        self.reader = vtkXArrayRectilinearSource()

        # List options
        self.state.update(
            {
                "xr_meta": {},
                "xr_vars": [],
                "available_datasets": [],
                "available_data_groups": [],
            }
        )
        self.state.available_catalogs = pan3d_catalogs.list_availables_search()

        # self.state.available_datasets = [
        #     {
        #         "id": "MITgcm_channel_flatbottom_02km_run01_phys-mon",
        #         "subtitle": "MITgcm channel simulations with flat bottom at 2km resolution physics field monthly mean climatology\n",
        #         "value": {
        #             "source": "pangeo",
        #             "id": "MITgcm_channel_flatbottom_02km_run01_phys-mon",
        #         },
        #     },
        # ]

        # Build UI
        self.ui = None
        self._build_ui()

    # -------------------------------------------------------------------------
    # Trame API
    # -------------------------------------------------------------------------

    def start(self, **kwargs):
        """Initialize the UI and start the server for GeoTrame."""
        self.ui.server.start(**kwargs)

    @property
    async def ready(self):
        """Coroutine to wait for the GeoTrame server to be ready."""
        await self.ui.ready

    @property
    def state(self):
        """Returns the current State of the Trame server."""
        return self.server.state

    @property
    def ctrl(self):
        """Returns the Controller for the Trame server."""
        return self.server.controller

    # -------------------------------------------------------------------------
    # GUI
    # -------------------------------------------------------------------------

    def _build_ui(self, *args, **kwargs):
        self.state.trame__title = "XArray Catalog Browser"
        with SinglePageWithDrawerLayout(self.server, full_height=True) as layout:
            self.ui = layout

            with layout.toolbar as tb:
                tb.density = "compact"
                layout.title.set_text("XArray Catalog Browser")
                v3.VSpacer()
                v3.VBtn(
                    icon="mdi-search-web",
                    click="ui_search_catalogs = !ui_search_catalogs",
                )
                v3.VProgressLinear(
                    indeterminate=("trame__busy",),
                    absolute=True,
                    location="bottom",
                    color="primary",
                    bg_color="rgba(0,0,0,0)",
                    opacity=0.2,
                    height=3,
                )

            CatalogSearch(
                update_catalog_search_term_function=self._update_catalog_search_term,
                catalog_search_function=self._catalog_search,
                catalog_term_search_function=self._catalog_term_option_search,
            )

            with layout.drawer as drawer:
                drawer.width = 350
                with v3.VToolbar(
                    density="compact",
                    flat=True,
                    color="white",
                    classes="pa-1 border-b-thin",
                ):
                    v3.VTextField(
                        placeholder="filter...",
                        v_model=("filter_query", ""),
                        density="compact",
                        hide_details=True,
                        prepend_inner_icon="mdi-magnify",
                        variant="outlined",
                        rounded=True,
                        clearable=True,
                    )
                    v3.VBtn(
                        icon="mdi-trash-can-outline",
                        flat=True,
                        density="compact",
                        classes="mx-1",
                        click="selected_dataset=[];available_datasets=[];",
                    )

                v3.VList(
                    mandatory=True,
                    selectable=True,
                    density="compact",
                    items=(
                        "available_datasets.filter((v) => !filter_query || (filter_query.length === 0) || v.subtitle.toLowerCase().includes(filter_query.toLowerCase()) || v.id.toLowerCase().includes(filter_query.toLowerCase()))",
                        TrameDefault(
                            available_datasets=[],
                            selected_dataset=[],
                        ),
                    ),
                    item_props=True,
                    item_value="value",
                    lines="three",
                    raw_attrs=['v-model:selected="selected_dataset"'],
                )

            with layout.content:
                with v3.VContainer(fluid=True, classes="full-height pa-0"):
                    with v3.VCard(flat=True, tile=True, v_if="selected_dataset.length"):
                        with v3.VCardTitle(
                            "{{ selected_dataset[0]?.id}}",
                            classes="d-flex",
                        ):
                            v3.VSpacer()
                            v3.VLabel("({{ selected_dataset[0]?.source }})")
                            v3.VBtn(
                                icon="mdi-cloud-download-outline",
                                density="compact",
                                click=self.save_data_origin,
                                flat=True,
                                classes="ml-2",
                            )
                            v3.VBtn(
                                icon="mdi-file-download-outline",
                                density="compact",
                                click=self.save_dataset,
                                flat=True,
                                classes="ml-2",
                            )
                        v3.VDivider()
                        v3.VSelect(
                            label="Coordinates",
                            v_model=("xr_coord", None),
                            items=("xr_coords", []),
                            density="compact",
                            hide_details=True,
                            flat=True,
                            variant="outlined",
                            classes="ma-2",
                        )

                        with v3.VTable(density="compact"):
                            with html.Tbody():
                                with html.Tr(
                                    v_for="attr, j in xr_meta[xr_coord] || []",
                                    key="j",
                                ):
                                    html.Td("{{ attr.key }}")
                                    html.Td("{{ attr.value }}")

                        v3.VSelect(
                            label="Data arrays",
                            v_model=("xr_array", None),
                            items=("xr_arrays", []),
                            density="compact",
                            hide_details=True,
                            flat=True,
                            variant="outlined",
                            classes="ma-2",
                        )

                        with v3.VTable(density="compact"):
                            with html.Tbody():
                                with html.Tr(
                                    v_for="attr, j in xr_meta[xr_array] || []",
                                    key="j",
                                ):
                                    html.Td("{{ attr.key }}")
                                    html.Td("{{ attr.value }}")
                        v3.VSelect(
                            v_if="xr_vars.length",
                            label="Variables",
                            v_model=("xr_var", None),
                            items=("xr_vars", []),
                            density="compact",
                            hide_details=True,
                            flat=True,
                            variant="outlined",
                            classes="ma-2",
                        )

                        with v3.VTable(density="compact", v_if="xr_vars.length"):
                            with html.Tbody():
                                with html.Tr(
                                    v_for="attr, j in xr_meta[xr_var] || []",
                                    key="j",
                                ):
                                    html.Td("{{ attr.key }}")
                                    html.Td("{{ attr.value }}")
                    v3.VCardText(
                        "Search dataset from the catalogs and select one from the list on the left to explore its content.",
                        v_else=True,
                    )

    # -------------------------------------------------------------------------
    # Triggers
    # -------------------------------------------------------------------------

    def _update_catalog_search_term(self, term_key, term_value):
        self.state.catalog_current_search[term_key] = term_value
        self.state.dirty("catalog_current_search")

    def _catalog_search(self):
        def load_results():
            catalog_id = self.state.catalog.get("id")
            results, group_name, message = pan3d_catalogs.search(
                catalog_id, **self.state.catalog_current_search
            )

            if len(results) > 0:
                self.state.available_data_groups.append(
                    {"name": group_name, "value": group_name}
                )
                self.state.available_datasets.extend(results)
                self.state.ui_catalog_search_message = message
                self.state.dirty("available_data_groups", "available_datasets")
            else:
                self.state.ui_catalog_search_message = (
                    "No results found for current search criteria."
                )

        self.run_as_async(
            load_results,
            loading_state="ui_catalog_term_search_loading",
            error_state="ui_catalog_search_message",
            unapplied_changes_state=None,
        )

    def _catalog_term_option_search(self):
        def load_terms():
            catalog_id = self.state.catalog.get("id")
            search_options = pan3d_catalogs.get_search_options(catalog_id)
            self.state.available_catalogs = [
                (
                    {
                        **catalog,
                        "search_terms": [
                            {"key": k, "options": v} for k, v in search_options.items()
                        ],
                    }
                    if catalog.get("id") == catalog_id
                    else catalog
                )
                for catalog in self.state.available_catalogs
            ]
            for catalog in self.state.available_catalogs:
                if catalog.get("id") == catalog_id:
                    self.state.catalog = catalog

        self.run_as_async(
            load_terms,
            loading_state="ui_catalog_term_search_loading",
            error_state="ui_catalog_search_message",
            unapplied_changes_state=None,
        )

    def run_as_async(
        self,
        function,
        loading_state="ui_loading",
        error_state="ui_error_message",
        unapplied_changes_state="ui_unapplied_changes",
    ):
        async def run():
            with self.state:
                if loading_state is not None:
                    self.state[loading_state] = True
                if error_state is not None:
                    self.state[error_state] = None
                if unapplied_changes_state is not None:
                    self.state[unapplied_changes_state] = False

            await asyncio.sleep(0.001)

            with self.state:
                try:
                    function()
                except Exception as e:
                    if error_state is not None:
                        self.state[error_state] = str(e)
                    else:
                        raise e
                if loading_state is not None:
                    self.state[loading_state] = False

            await asyncio.sleep(0.001)

        if self.current_event_loop.is_running():
            asyncio.run_coroutine_threadsafe(run(), self.current_event_loop)
        else:
            # Pytest environment needs synchronous execution
            function()

    def save_data_origin(self):
        selected_dataset = self.state.selected_dataset
        if not selected_dataset:
            return

        file = Path(f"{selected_dataset[0].get('id')}.data-origin.json")
        file.write_text(json.dumps({"data_origin": selected_dataset[0]}))

    def save_dataset(self):
        selected_dataset = self.state.selected_dataset
        if not selected_dataset:
            return

        name = f"{selected_dataset[0].get('id')}.nc"

        self.reader.input.to_netcdf(name)

    # -----------------------------------------------------
    # State change callbacks
    # -----------------------------------------------------

    @change("ui_search_catalogs")
    def _on_change_ui_search_catalogs(self, ui_search_catalogs, **kwargs):
        if ui_search_catalogs:
            self.state.catalog = self.state.available_catalogs[0]
        else:
            self.state.catalog = None

    @change("catalog")
    def _on_change_catalog(self, **_):
        self.state.catalog_current_search = {}
        self.state.ui_catalog_search_message = None

    @change("selected_dataset")
    def on_selection_change(self, selected_dataset, **_):
        if not selected_dataset:
            self.state.xarray_info = []
            return

        # Load metadata
        self.reader.load({"data_origin": selected_dataset[0]})
        xr = self.reader.input
        available_arrays = self.reader.available_arrays

        # Extract metadata
        xr_meta = {}
        xr_coords = []
        xr_arrays = []
        xr_vars = []

        coords = set(xr.coords.keys())
        data = set(available_arrays or [])
        for name in xr.variables:
            attrs = xr_meta.setdefault(name, [])
            attrs.extend(
                [{"key": str(k), "value": str(v)} for k, v in xr[name].attrs.items()]
            )
            attrs.insert(
                0,
                {"key": "type", "value": str(xr[name].dtype)},
            )
            attrs.insert(
                0,
                {
                    "key": "shape",
                    "value": list(xr[name].shape),
                },
            )
            if name in coords:
                xr_coords.append(name)
                length = xr[name].size
                shape = xr[name].shape
                if length > 1 and len(shape) == 1:
                    attrs.insert(
                        0,
                        {
                            "key": "range",
                            "value": f"[{xr[name].values[0]}, {xr[name].values[-1]}]",
                        },
                    )

            elif name in data:
                attrs.insert(
                    0,
                    {
                        "key": "dimensions",
                        "value": f"({','.join(xr[name].dims)})",
                    },
                )
                xr_arrays.append(name)
            else:
                xr_vars.append(name)

        # Update UI
        self.state.xr_meta = xr_meta
        self.state.xr_coords = xr_coords
        self.state.xr_arrays = xr_arrays
        self.state.xr_vars = xr_vars
        self.state.xr_coord = xr_coords[0] if len(xr_coords) else ""
        self.state.xr_array = xr_arrays[0] if len(xr_arrays) else ""
        self.state.xr_var = xr_vars[0] if len(xr_vars) else ""


# -----------------------------------------------------------------------------
# Main executable
# -----------------------------------------------------------------------------


def main():
    app = CatalogBrowser()
    app.start()


if __name__ == "__main__":
    main()



import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
from vtkmodules.vtkCommonCore import vtkObject
from vtkmodules.vtkFiltersGeometry import vtkDataSetSurfaceFilter

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import (
    vtkInteractorStyleSwitch,  # noqa: F401
    vtkInteractorStyleTerrain,
)
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkRenderingAnnotation import vtkAxesActor
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
)

from pan3d.filters.globe import ProjectToSphere
from pan3d.ui.globe import GlobeRenderingSettings
from pan3d.ui.vtk_view import Pan3DView
from pan3d.utils.common import ControlPanel, Explorer, SummaryToolbar
from pan3d.utils.globe import get_continent_outlines, get_globe, get_globe_textures
from pan3d.widgets.scalar_bar import ScalarBar
from pan3d.xarray.algorithm import vtkXArrayRectilinearSource
from trame.decorators import change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import vuetify3 as v3

# Prevent view-up warning
vtkObject.GlobalWarningDisplayOff()


class GlobeExplorer(Explorer):
    """
    A Trame based pan3D explorer to visualize 3D geographic data projected onto a globe
    representing the earth or projected using various cartographic projections.
    The prerequisite is that the coordinates of the dataset need to be in lat-long format.
    This explorer uses the pan3D DatasetBuilder and it's operability with xarray to fetch
    relevant data and visualizes it using VTK while interacting with the slice in 2D or 3D.
    """

    def __init__(self, xarray=None, source=None, server=None, local_rendering=None):
        super().__init__(xarray, source, server, local_rendering)
        if self.source is None:
            self.source = vtkXArrayRectilinearSource()  # To initialize the pipeline
        self.textures = get_globe_textures()
        self.state.textures = list(self.textures.keys())

        self._setup_vtk()
        self._build_ui()

    # -------------------------------------------------------------------------
    # VTK Setup
    # -------------------------------------------------------------------------

    def _setup_vtk(self):
        self.renderer = vtkRenderer(background=(0.8, 0.8, 0.8))
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow(off_screen_rendering=1)

        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        self.interactor.SetInteractorStyle(vtkInteractorStyleTerrain())

        self.globe = get_globe()
        self.gmapper = vtkPolyDataMapper(input_data_object=self.globe)
        self.gactor = vtkActor(mapper=self.gmapper, visibility=1)

        self.continents = get_continent_outlines()
        self.cmapper = vtkPolyDataMapper(input_data_object=self.continents)
        self.cactor = vtkActor(mapper=self.cmapper, visibility=1)

        dglobe = ProjectToSphere()
        dglobe.isData = True
        dglobe.input_connection = self.source.output_port
        self.dglobe = dglobe
        # Need explicit geometry extraction when used with WASM
        self.geometry = vtkDataSetSurfaceFilter(
            input_connection=self.dglobe.output_port
        )

        self.mapper = vtkPolyDataMapper(input_connection=self.geometry.output_port)
        self.actor = vtkActor(mapper=self.mapper, visibility=0)

        # Camera
        camera = self.renderer.GetActiveCamera()
        camera.SetFocalPoint(0, 0, 0)
        camera.SetPosition(0, -1, 0)
        camera.SetViewUp(0, 0, 1)
        self.renderer.ResetCamera()

        self.interactor.Initialize()

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

    # -------------------------------------------------------------------------
    # UI
    # -------------------------------------------------------------------------

    def _build_ui(self, **kwargs):
        self.state.update(
            {
                "trame__title": "Globe Viewer",
                "slice_extents": {},
                "axis_names": [],
                "t_labels": [],
                "max_time_width": 0,
                "max_time_index_width": 0,
                "dataset_bounds": [0, 1, 0, 1, 0, 1],
                "render_shadow": False,
            }
        )
        with VAppLayout(self.server, fill_height=True) as layout:
            self.ui = layout

            # 3D view
            Pan3DView(
                self.render_window,
                local_rendering=self.local_rendering,
                widgets=[self.widget],
                disable_style_toggle=True,
                disable_roll=True,
                disable_axis_align=True,
            )

            # Scalar bar
            ScalarBar(
                ctx_name="scalar_bar",
                v_show="!control_expended",
                v_if="color_by",
            )

            # Save dialog
            with v3.VDialog(v_model=("show_save_dialog", False)):
                with v3.VCard(classes="mx-auto w-50"):
                    v3.VCardTitle("Save dataset to disk")
                    v3.VDivider()
                    with v3.VCardText():
                        v3.VTextField(
                            label="File path to save",
                            v_model=("save_dataset_path", ""),
                            hide_details=True,
                        )
                    with v3.VCardActions():
                        v3.VSpacer()
                        v3.VBtn(
                            "Save",
                            classes="text-none",
                            variant="flat",
                            color="primary",
                            click=(self.save_dataset, "[save_dataset_path]"),
                        )
                        v3.VBtn(
                            "Cancel",
                            classes="text-none",
                            variant="flat",
                            click="show_save_dialog=false",
                        )

            # Error messages
            v3.VAlert(
                v_if=("data_origin_error", False),
                border="start",
                max_width=700,
                rounded="lg",
                text=("data_origin_error", ""),
                title="Failed to load data",
                type="error",
                variant="tonal",
                style="position:absolute;bottom:1rem;right:1rem;",
            )

            # Summary toolbar
            SummaryToolbar(
                v_show="!control_expended",
                v_if="slice_t_max > 0",
            )

            with ControlPanel(
                enable_data_selection=(self.xarray is None),
                source=self.source,
                toggle="control_expended",
                load_dataset=self.load_dataset,
                import_file_upload=self.import_file_upload,
                export_file_download=self.export_state,
                xr_update_info="xr_update_info",
                panel_label="Globe Explorer",
            ).ui_content:
                GlobeRenderingSettings(
                    ctx_name="rendering",
                    source=self.source,
                    update_rendering=self.update_rendering,
                )

    # -----------------------------------------------------
    # State change callbacks
    # -----------------------------------------------------

    @change("opacity", "representation", "cell_size", "render_shadow")
    def _on_change_opacity(
        self, representation, opacity, cell_size, render_shadow, **_
    ):
        property = self.actor.property
        property.render_lines_as_tubes = render_shadow
        property.render_points_as_spheres = render_shadow
        property.line_width = cell_size
        property.point_size = cell_size
        property.opacity = float(opacity) if representation == 2 else 1
        property.representation = representation

        self.ctrl.view_update()

    @change("bump_radius")
    def _on_bump_radius_change(self, bump_radius, **_):
        self.dglobe.bump_radius = bump_radius
        self.ctrl.view_update()

    @change("texture")
    def _on_texture_preset(self, texture, **_):
        self.gactor.SetTexture(self.textures[texture])
        self.ctrl.view_update()

    def update_rendering(self, reset_camera=False):
        self.state.dirty_data = False

        self.gactor.visibility = 1
        self.gactor.SetTexture(self.textures[self.state.texture])
        self.renderer.AddActor(self.gactor)

        self.cactor.visibility = 1
        self.cactor.GetProperty().SetRepresentationToWireframe()
        self.cactor.GetProperty().SetColor(1.0, 1.0, 1.0)
        self.renderer.AddActor(self.cactor)

        if self.actor.visibility == 0:
            self.actor.visibility = 1
            self.renderer.AddActor(self.actor)
            self.renderer.ResetCamera()
            if self.ctrl.view_update_force.exists():
                self.ctrl.view_update_force(push_camera=True)

        if reset_camera:
            self.ctrl.view_reset_camera()
        else:
            self.ctrl.view_update()


# -----------------------------------------------------------------------------
# Main executable
# -----------------------------------------------------------------------------


def main():
    app = GlobeExplorer()
    app.start()


if __name__ == "__main__":
    main()



from typing import Optional

import numpy as np
import xarray as xr
from vtkmodules.vtkCommonDataModel import vtkDataSet

from pan3d.xarray import algorithm, datasets


class _LocIndexer:
    def __init__(self, parent: "VTKAccessor"):
        self.parent = parent

    def __getitem__(self, key) -> xr.DataArray:
        return self.parent._xarray.loc[key]

    def __setitem__(self, key, value) -> None:
        self.parent._xarray.__setitem__(self, key, value)


@xr.register_dataset_accessor("pan3d")
class Pan3DAccessor:
    accessor_id = 0

    @classmethod
    def next_id(cls):
        cls.accessor_id += 1
        return f"pan3d_accessor_{cls.accessor_id}"

    def __init__(self, xarray):
        self.xarray = xarray
        self.accessor_id = 0
        self._local_rendering = None
        self._viewer_preview = None
        self._viewer_slicer = None
        self._viewer_globe = None
        self._viewer_contour = None

    @property
    def local(self):
        """Builder pattern for viewer creation using local rendering"""
        self._local_rendering = "wasm"
        return self

    @property
    def remote(self):
        """Builder pattern for viewer creation using remote rendering"""
        self._local_rendering = None
        return self

    @property
    def preview(self):
        from pan3d.viewers.preview import XArrayViewer

        if self._viewer_preview is None:
            self._viewer_preview = XArrayViewer(
                xarray=self.xarray,
                server=self.next_id(),
                local_rendering=self._local_rendering,
            )

        return self._viewer_preview

    @property
    def slicer(self):
        from pan3d.explorers.slicer import SliceExplorer

        if self._viewer_slicer is None:
            self._viewer_slicer = SliceExplorer(
                xarray=self.xarray,
                server=self.next_id(),
                local_rendering=self._local_rendering,
            )

        return self._viewer_slicer

    @property
    def globe(self):
        from pan3d.explorers.globe import GlobeExplorer

        if self._viewer_globe is None:
            self._viewer_globe = GlobeExplorer(
                xarray=self.xarray,
                server=self.next_id(),
                local_rendering=self._local_rendering,
            )

        return self._viewer_globe

    @property
    def contour(self):
        from pan3d.explorers.contour import ContourExplorer

        if self._viewer_contour is None:
            self._viewer_contour = ContourExplorer(
                xarray=self.xarray,
                server=self.next_id(),
                local_rendering=self._local_rendering,
            )

        return self._viewer_contour


@xr.register_dataarray_accessor("vtk")
class VTKAccessor:
    def __init__(self, xarray_obj: xr.DataArray):
        self._xarray = xarray_obj

    def __getitem__(self, key):
        return self._xarray.__getitem__(key)

    @property
    def data(self):
        return self._xarray.values

    @property
    def loc(self) -> _LocIndexer:
        """Attribute for location based indexing like pandas."""
        return _LocIndexer(self)

    def _get_array(self, key, scale=1):
        try:
            values = self._xarray[key].values
            if "float" not in str(values.dtype) and "int" not in str(values.dtype):
                # non-numeric coordinate, assign array of scaled indices
                values = np.array(range(len(values))) * scale

            return values
        except KeyError as err:
            msg = f"Key {key} not present in DataArray. Choices are: {list(self._xarray.coords.keys())}"
            raise KeyError(msg) from err

    def dataset(
        self,
        x: Optional[str] = None,
        y: Optional[str] = None,
        z: Optional[str] = None,
        order: Optional[str] = None,
        component: Optional[str] = None,
        mesh_type: Optional[str] = None,
        scales: Optional[dict] = None,
    ) -> vtkDataSet:
        if mesh_type is None:  # Try to guess mesh type
            max_ndim = max(
                *[self._get_array(n).ndim for n in (x, y, z) if n is not None]
            )
            mesh_type = "structured" if max_ndim > 1 else "rectilinear"

        try:
            builder = getattr(datasets, mesh_type)
        except KeyError as err:
            raise KeyError() from err
        return builder(
            self, x=x, y=y, z=z, order=order, component=component, scales=scales
        )

    def algorithm(
        self,
        x: Optional[str] = None,
        y: Optional[str] = None,
        z: Optional[str] = None,
        t: Optional[str] = None,
        arrays: Optional[list[str]] = None,
        order: str = "C",
    ):
        return algorithm.vtkXArrayRectilinearSource(
            input=self._xarray,
            x=x,
            y=y,
            z=z,
            t=t,
            arrays=arrays,
            order=order,
        )



import warnings
from typing import Optional

import numpy as np
from vtkmodules.vtkCommonDataModel import vtkRectilinearGrid, vtkStructuredGrid

from pan3d.xarray.errors import DataCopyWarning


def imagedata_to_rectilinear(image_data):
    output = vtkRectilinearGrid()
    output.SetDimensions(image_data.dimensions)
    origin = image_data.origin
    spacing = image_data.spacing
    extent = image_data.extent
    coords = []
    for axis in range(3):
        coords.append(
            np.array(
                [
                    origin[axis] + (spacing[axis] * i)
                    for i in range(extent[axis * 2], extent[axis * 2 + 1] + 1)
                ],
                dtype=np.double,
            )
        )

    output.x_coordinates, output.y_coordinates, output.z_coordinates = coords
    output.point_data.ShallowCopy(image_data.point_data)
    output.cell_data.ShallowCopy(image_data.cell_data)
    output.field_data.ShallowCopy(image_data.field_data)

    return output


def _coerce_shapes(*arrs):
    """Coerce all argument arrays to have the same shape."""
    maxi = 0
    ndim = 0
    for i, arr in enumerate(arrs):
        if arr is None:
            continue
        if arr.ndim > ndim:
            ndim = arr.ndim
            maxi = i
    # print(arrs)
    # if ndim != len(arrs) - (*arrs,).count(None):
    #     print(ndim, len(arrs))
    #     raise ValueError
    if ndim < 1:
        raise ValueError
    shape = arrs[maxi].shape
    reshaped = []
    for arr in arrs:
        if arr is not None and arr.shape != shape:
            if arr.ndim < ndim:
                reshaped.append(np.repeat([arr], shape[2 - maxi], axis=2 - maxi))
            else:
                raise ValueError
        else:
            reshaped.append(arr)
    return reshaped


def _points(
    accessor,
    x: Optional[str] = None,
    y: Optional[str] = None,
    z: Optional[str] = None,
    order: Optional[str] = "F",
    scales: Optional[dict] = None,
):
    """Generate structured points as new array."""
    if order is None:
        order = "F"
    ndim = 3 - (x, y, z).count(None)
    if ndim < 2:
        if ndim == 1:
            raise ValueError(
                "One dimensional structured grids should be rectilinear grids."
            )
        raise ValueError("You must specify at least two dimensions as X, Y, or Z.")
    if x is not None:
        x = accessor._get_array(x, scale=(scales and scales.get(x)) or 1)
    if y is not None:
        y = accessor._get_array(y, scale=(scales and scales.get(y)) or 1)
    if z is not None:
        z = accessor._get_array(z, scale=(scales and scales.get(z)) or 1)
    arrs = _coerce_shapes(x, y, z)
    x, y, z = arrs
    arr = next(a for a in arrs if a is not None)
    points = np.zeros((arr.size, 3), dtype=arr.dtype)
    if x is not None:
        points[:, 0] = x.ravel(order=order)
    if y is not None:
        points[:, 1] = y.ravel(order=order)
    if z is not None:
        points[:, 2] = z.ravel(order=order)
    shape = list(x.shape) + [1] * (3 - ndim)
    return points, shape


def rectilinear(
    accessor,
    x: Optional[str] = None,
    y: Optional[str] = None,
    z: Optional[str] = None,
    order: Optional[str] = "C",
    component: Optional[str] = None,
    scales: Optional[dict] = None,
):
    if scales is None:
        scales = {}

    ndim = 3 - (x, y, z).count(None)
    if ndim < 1:
        raise ValueError("You must specify at least one dimension as X, Y, or Z.")

    # Build dataset
    dataset = vtkRectilinearGrid()

    if x is not None:
        dataset.x_coordinates = accessor._get_array(x, scale=scales.get(x, 1))
    if y is not None:
        dataset.y_coordinates = accessor._get_array(y, scale=scales.get(y, 1))
    if z is not None:
        dataset.z_coordinates = accessor._get_array(z, scale=scales.get(z, 1))

    # Update grid size
    dataset.dimensions = [
        dataset.x_coordinates.size,
        dataset.y_coordinates.size,
        dataset.z_coordinates.size,
    ]

    # Handle field
    values = accessor.data
    values_dim = values.ndim
    if component is not None:
        # if ndim < values.ndim and values.ndim == ndim + 1:
        # Assuming additional component array
        dims = set(accessor._xarray.dims)
        dims.discard(component)
        print("values changed - transpose")
        values = accessor._xarray.transpose(
            *dims, component, transpose_coords=True
        ).values
        values = values.reshape((-1, values.shape[-1]), order=order)
        warnings.warn(
            DataCopyWarning(
                "Made a copy of the multicomponent array - VTK data not shared with xarray."
            ),
            stacklevel=2,
        )
        ndim += 1
    else:
        print("values changed - ravel")
        values = values.ravel(order=order)

    # Validate dimensions of field
    if values_dim != ndim:
        msg = f"Dimensional mismatch between specified X, Y, Z coords and dimensionality of DataArray ({ndim} vs {values_dim})"
        if ndim > values_dim:
            raise ValueError(
                "".join(
                    [
                        msg,
                        ". Too many coordinate dimensions specified leave out Y and/or Z.",
                    ]
                )
            )
        raise ValueError(
            "".join(
                [
                    msg,
                    ". Too few coordinate dimensions specified. Be sure to specify Y and/or Z or reduce the dimensionality of the DataArray by indexing along non-spatial coordinates like Time.",
                ]
            )
        )

    array_name = str(accessor._xarray.name or "data")
    dataset.point_data[array_name] = values
    return dataset


def structured(
    accessor,
    x: Optional[str] = None,
    y: Optional[str] = None,
    z: Optional[str] = None,
    order: Optional[str] = "F",
    component: Optional[str] = None,
    scales: Optional[dict] = None,
):
    if scales is None:
        scales = {}

    points, shape = _points(accessor, x=x, y=y, z=z, order=order, scales=scales)

    dataset = vtkStructuredGrid()
    dataset.SetDimensions(shape)
    dataset.points = points
    dataset.point_data[accessor._xarray.name or "data"] = accessor.data.ravel(
        order=order
    )

    return dataset



import vtkmodules.vtkRenderingOpenGL2  # noqa: F401
from vtkmodules.vtkCommonDataModel import (
    vtkPlane,
)
from vtkmodules.vtkFiltersCore import (
    vtkCutter,
)
from vtkmodules.vtkFiltersModeling import (
    vtkOutlineFilter,
)

# VTK factory initialization
from vtkmodules.vtkInteractionStyle import vtkInteractorStyleSwitch  # noqa: F401
from vtkmodules.vtkInteractionWidgets import vtkOrientationMarkerWidget
from vtkmodules.vtkRenderingAnnotation import (
    vtkAxesActor,
)
from vtkmodules.vtkRenderingCore import (
    vtkActor,
    vtkDataSetMapper,
    vtkPolyDataMapper,
    vtkRenderer,
    vtkRenderWindow,
    vtkRenderWindowInteractor,
)

from pan3d.ui.slicer import SliceRenderingSettings
from pan3d.ui.vtk_view import Pan3DView
from pan3d.utils.common import ControlPanel, Explorer, SummaryToolbar
from pan3d.widgets.scalar_bar import ScalarBar
from pan3d.xarray.algorithm import vtkXArrayRectilinearSource
from trame.decorators import change
from trame.ui.vuetify3 import VAppLayout
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class Pan3DSlicerView(Pan3DView):
    def __init__(self, render_window, **kwargs):
        super().__init__(render_window=render_window, **kwargs)
        with self.toolbar:
            v3.VDivider()
            with v3.VTooltip(text="Slice View Mode (2D/3D)"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VCheckbox(
                        v_bind="props",
                        v_model=("view_mode", "3D"),
                        true_icon="mdi-video-2d",
                        false_icon="mdi-video-3d",
                        density="compact",
                        hide_details=True,
                        true_value="2D",
                        false_value="3D",
                    )
            v3.VDivider()
            with v3.VTooltip(text="Outline 3D Data Extents"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VCheckbox(
                        v_bind="props",
                        v_model=("outline", True),
                        true_icon="mdi-cube-outline",
                        false_icon="mdi-cube-outline",
                        density="compact",
                        hide_details=True,
                        disabled=("view_mode === '2D'",),
                    )
            with v3.VTooltip(text="Use Transparency for 3D Data"):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VCheckbox(
                        v_bind="props",
                        v_model=("tdata", True),
                        true_icon="mdi-texture-box",
                        false_icon="mdi-texture-box",
                        density="compact",
                        hide_details=True,
                        disabled=("view_mode === '2D'",),
                    )


class SliceSummary(html.Div):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        with self:
            html.Div(
                "{{slice_axis}}",
                classes="text-subtitle-1 text-capitalize text-left",
                style="transform-origin: 50% 50%; transform: rotate(-90deg) translateX(-100%) translateY(-1rem); position: absolute;",
            )
            html.Div(
                "{{parseFloat(bounds[slice_axes.indexOf(slice_axis)*2 + 1]).toFixed(2)}}",
                classes="text-subtitle-1 mx-1",
            )
            v3.VSlider(
                v_show="slice_axis === slice_axes[0]",
                thumb_label="always",
                thumb_size=16,
                style="pointer-events: auto;",
                hide_details=True,
                classes="flex-fill",
                direction="vertical",
                v_model=("cut_x",),
                min=("bounds[0]",),
                max=("bounds[1]",),
            )
            v3.VSlider(
                v_show="slice_axis === slice_axes[1]",
                thumb_label="always",
                thumb_size=16,
                style="pointer-events: auto;",
                hide_details=True,
                classes="flex-fill",
                direction="vertical",
                v_model=("cut_y",),
                min=("bounds[2]",),
                max=("bounds[3]",),
            )
            v3.VSlider(
                v_show="slice_axis === slice_axes[2]",
                thumb_label="always",
                thumb_size=16,
                style="pointer-events: auto;",
                hide_details=True,
                classes="flex-fill",
                direction="vertical",
                v_model=("cut_z",),
                min=("bounds[4]",),
                max=("bounds[5]",),
            )

            html.Div(
                "{{parseFloat(bounds[slice_axes.indexOf(slice_axis)*2]).toFixed(2)}}",
                classes="text-subtitle-1 mx-1",
            )


class SliceExplorer(Explorer):
    """
    A Trame based pan3D explorer to visualize 3D using slices along different dimensions

    This explorer uses the pan3D DatasetBuilder and it's operability with xarray to fetch
    relevant data and allows users to specify a specific slice of interest and visualize it
    using VTK while interacting with the slice in 2D or 3D.
    """

    def __init__(self, xarray=None, source=None, server=None, local_rendering=None):
        super().__init__(xarray, source, server, local_rendering)
        if self.source is None:
            self.source = vtkXArrayRectilinearSource()  # To initialize the pipeline

        self._setup_vtk()
        self._build_ui()

    def _setup_vtk(self):
        ds = self.source()
        bounds = ds.bounds
        self.normal = [0, 0, 1]
        self.origin = [
            0.5 * (bounds[0] + bounds[1]),
            0.5 * (bounds[2] + bounds[3]),
            0.5 * (bounds[4] + bounds[5]),
        ]

        # Build rendering pipeline
        self.renderer = vtkRenderer()
        self.interactor = vtkRenderWindowInteractor()
        self.render_window = vtkRenderWindow()

        plane = vtkPlane()
        plane.SetOrigin(self.origin)
        plane.SetNormal(self.normal)
        cutter = vtkCutter()
        cutter.SetCutFunction(plane)
        cutter.input_connection = self.source.output_port
        slice_actor = vtkActor()
        slice_mapper = vtkDataSetMapper()
        slice_mapper.SetInputConnection(cutter.GetOutputPort())
        slice_mapper.SetScalarModeToUsePointFieldData()
        slice_mapper.InterpolateScalarsBeforeMappingOn()
        slice_actor.SetMapper(slice_mapper)
        self.plane = plane
        self.cutter = cutter
        self.slice_actor = slice_actor
        self.mapper = slice_mapper

        outline = vtkOutlineFilter()
        outline_actor = vtkActor()
        outline_mapper = vtkPolyDataMapper()
        outline.input_connection = self.source.output_port
        outline_mapper.SetInputConnection(outline.GetOutputPort())
        outline_actor.SetMapper(outline_mapper)
        outline_actor.GetProperty().SetColor(0.5, 0.5, 0.5)
        self.outline = outline
        self.outline_actor = outline_actor
        self.outline_mapper = outline_mapper

        data_actor = vtkActor()
        data_mapper = vtkDataSetMapper()
        data_mapper.input_connection = self.source.output_port
        data_actor.SetMapper(data_mapper)
        data_actor.GetProperty().SetOpacity(0.1)
        data_actor.SetVisibility(False)
        self.data_actor = data_actor
        self.data_mapper = data_mapper

        self.renderer.SetBackground(1.0, 1.0, 1.0)
        self.render_window.OffScreenRenderingOn()
        self.render_window.AddRenderer(self.renderer)
        self.interactor.SetRenderWindow(self.render_window)
        self.interactor.GetInteractorStyle().SetCurrentStyleToTrackballCamera()

        self.renderer.AddActor(self.outline_actor)
        self.renderer.AddActor(self.data_actor)
        self.renderer.AddActor(self.slice_actor)

        axes_actor = vtkAxesActor()
        self.widget = vtkOrientationMarkerWidget()
        self.widget.SetOrientationMarker(axes_actor)
        self.widget.SetInteractor(self.interactor)
        self.widget.SetViewport(0.85, 0, 1, 0.15)
        self.widget.EnabledOn()
        self.widget.InteractiveOff()

    # -------------------------------------------------------------------------
    # GUI definition
    # -------------------------------------------------------------------------

    def _build_ui(self, *args, **kwargs):
        self.state.update(
            {
                "trame__title": "Slice Explorer",
                "import_pending": False,
                "control_expended": True,
                "axis_names": ["X", "Y", "Z"],
                "scale_x": 1,
                "scale_y": 1,
                "scale_z": 1,
                "cut_x": 0.5,
                "cut_y": 0.5,
                "cut_z": 0.5,
                "bounds": [0.0, 1.0, 0.0, 1.0, 0.0, 1.0],
                "slice_axis": "Z",
            }
        )
        with VAppLayout(self.server, fill_height=True) as layout:
            self.ui = layout

            # 3D view
            Pan3DSlicerView(
                self.render_window,
                local_rendering=self.local_rendering,
                widgets=[self.widget],
            )

            # Scalar bar
            ScalarBar(
                ctx_name="scalar_bar",
                v_show="!control_expended",
                v_if="color_by",
            )

            # Save dialog
            with v3.VDialog(v_model=("show_save_dialog", False)):
                with v3.VCard(classes="mx-auto w-50"):
                    v3.VCardTitle("Save dataset to disk")
                    v3.VDivider()
                    with v3.VCardText():
                        v3.VTextField(
                            label="File path to save",
                            v_model=("save_dataset_path", ""),
                            hide_details=True,
                        )
                    with v3.VCardActions():
                        v3.VSpacer()
                        v3.VBtn(
                            "Save",
                            classes="text-none",
                            variant="flat",
                            color="primary",
                            click=(self.save_dataset, "[save_dataset_path]"),
                        )
                        v3.VBtn(
                            "Cancel",
                            classes="text-none",
                            variant="flat",
                            click="show_save_dialog=false",
                        )

            # Error messages
            v3.VAlert(
                v_if=("data_origin_error", False),
                border="start",
                max_width=700,
                rounded="lg",
                text=("data_origin_error", ""),
                title="Failed to load data",
                type="error",
                variant="tonal",
                style="position:absolute;bottom:1rem;right:1rem;",
            )

            # Summary toolbar
            SummaryToolbar(
                v_show="!control_expended",
                v_if="slice_t_max > 0",
            )

            # Sliders overlay
            SliceSummary(
                v_if="!control_expended",
                classes="d-flex align-center flex-column",
                style="position: absolute; left: 0; top: 10%; bottom: 10%; z-index: 2; pointer-events: none; min-width: 5rem;",
            )

            # Control panel
            with ControlPanel(
                enable_data_selection=(self.xarray is None),
                source=self.source,
                toggle="control_expended",
                load_dataset=self.load_dataset,
                import_file_upload=self.import_file_upload,
                export_file_download=self.export_state,
                xr_update_info="xr_update_info",
                panel_label="Slice Explorer",
            ).ui_content:
                SliceRenderingSettings(
                    ctx_name="rendering",
                    source=self.source,
                    update_rendering=self.update_rendering,
                )

    def update_rendering(self, reset_camera=False):
        self.renderer.ResetCamera()

        if self.local_rendering:
            self.ctrl.view_update(push_camera=True)

        self.ctrl.view_reset_camera()

    # -------------------------------------------------------------------------
    # Property API
    # -------------------------------------------------------------------------

    @property
    def slice_axis(self):
        """
        Returns the active axis along which the slice is performed
        """
        return self.state.slice_axis

    @slice_axis.setter
    def slice_axis(self, axis: str) -> None:
        """
        Sets the active axis along which the slice is performed
        """
        with self.state:
            self.state.slice_axis = axis

    @property
    def slice_value(self):
        """
        Returns the value(origin) for the dimension along which the slice
        is performed
        """
        s = self.state
        axis = "xyz"[s.slice_axes.index(s.slice_axis)]
        return s[f"cut_{axis}"]

    @slice_value.setter
    def slice_value(self, value: float) -> None:
        """
        Sets the value(origin) for the dimension along which the slice
        is performed
        """
        with self.state:
            s = self.state
            axis = "xyz"[s.slice_axes.index(s.slice_axis)]
            s[f"cut_{axis}"] = value

    @property
    def view_mode(self):
        """
        Returns the interaction mode (2D/3D) for the slice
        """
        return self.state.view_mode

    @view_mode.setter
    def view_mode(self, mode):
        """
        Sets the interaction mode (2D/3D) for the slice,
        and updates camera accordingly. Uses isometric view for 3D
        """
        with self.state:
            self.state.view_mode = mode

    @property
    def scale_axis(self):
        s = self.state
        return [s.x_scale, s.y_scale, s.z_scale]

    @scale_axis.setter
    def scale_axis(self, sfac):
        s = self.state
        s.x_scale = float(sfac[0])
        s.y_scale = float(sfac[1])
        s.z_scale = float(sfac[2])
        self.slice_actor.SetScale(*sfac)
        self.data_actor.SetScale(*sfac)
        self.outline_actor.SetScale(*sfac)
        self.on_view_mode_change(s.view_mode)

    # -------------------------------------------------------------------------
    # UI triggers
    # -------------------------------------------------------------------------

    @change("scale_x", "scale_y", "scale_z")
    def _on_scale_change(self, scale_x, scale_y, scale_z, **_):
        scales = [float(scale_x), float(scale_y), float(scale_z)]
        self.slice_actor.SetScale(*scales)
        self.data_actor.SetScale(*scales)
        self.outline_actor.SetScale(*scales)

        self.renderer.ResetCamera()

        self.on_view_mode_change(self.state.view_mode)

    def _set_view_2D(self, axis):
        camera = self.renderer.GetActiveCamera()
        view_up = [0, 0, 1] if axis == 1 else [0, 1, 0]
        direction = [0, 0, 0]
        direction[axis] = 1
        camera.SetFocalPoint(0, 0, 0)
        camera.SetPosition(*direction)
        camera.SetViewUp(*view_up)
        camera.OrthogonalizeViewUp()

        self.outline_actor.SetVisibility(False)
        self.data_actor.SetVisibility(False)

        self.renderer.ResetCamera()
        self.ctrl.view_update()

    def _set_view_3D(self):
        if self.state.outline:
            self.outline_actor.SetVisibility(True)
        if self.state.tdata:
            self.data_actor.SetVisibility(True)

        self.renderer.ResetCamera()
        self.ctrl.view_update()

    @change("view_mode")
    def on_view_mode_change(self, view_mode, **_):
        """
        Performs all the steps necessary when user toggles the view mode
        """
        if view_mode == "3D":
            self._set_view_3D()
        elif view_mode == "2D":
            s = self.state
            axis_idx = s.axis_names.index(s.slice_axis)
            self._set_view_2D(axis_idx)

    @change("slice_axis", "slice_t", "cut_x", "cut_y", "cut_z")
    def _on_data_change(
        self, slice_axis, axis_names, slice_t, cut_x, cut_y, cut_z, **_
    ):
        """
        Performs all the steps necessary when the user updates any properties
        that requires a new data update. E.g. changing the data variable for
        visualization, or changing active time, or changing slice value.
        """
        normal = [0, 0, 0]
        normal[axis_names.index(slice_axis)] = 1
        origin = [
            float(cut_x),
            float(cut_y),
            float(cut_z),
        ]

        self.source.t_index = slice_t
        self.plane.SetOrigin(origin)
        self.plane.SetNormal(normal)

        if self.state.view_mode == "2D":
            self.on_view_mode_change("2D")
            self.renderer.ResetCamera()

        self.ctrl.view_update()

    @change("outline", "tdata")
    def _on_rep_change(self, outline, tdata, **_):
        """
        Performs all the steps necessary when user specifies 3D interaction options
        """
        self.outline_actor.SetVisibility(outline)
        self.data_actor.SetVisibility(tdata)
        self.ctrl.view_update()


def main():
    app = SliceExplorer()
    app.start()


if __name__ == "__main__":
    main()



import json
from pathlib import Path

import numpy as np
from vtkmodules.vtkCommonCore import vtkLookupTable
from vtkmodules.vtkRenderingCore import vtkColorTransferFunction

PRESETS = {
    item.get("Name"): item
    for item in json.loads(Path(__file__).with_name("presets.json").read_text())
}

LUTS = {}


def update_range(lut: vtkColorTransferFunction, data_range):
    prev_min, prev_max = lut.GetRange()
    prev_delta = prev_max - prev_min

    if prev_delta < 0.001:
        return

    node = [0, 0, 0, 0, 0, 0]
    next_delta = data_range[1] - data_range[0]
    next_nodes = []

    for i in range(lut.GetSize()):
        lut.GetNodeValue(i, node)
        node[0] = next_delta * (node[0] - prev_min) / prev_delta + data_range[0]
        next_nodes.append(list(node))

    lut.RemoveAllPoints()
    for n in next_nodes:
        lut.AddRGBPoint(*n)


def get_preset(preset_name: str) -> vtkColorTransferFunction:
    if preset_name in LUTS:
        return LUTS[preset_name]

    lut = LUTS.setdefault(preset_name, vtkColorTransferFunction())
    preset = PRESETS[preset_name]
    srgb = np.array(preset["RGBPoints"])
    color_space = preset["ColorSpace"]

    if color_space == "Diverging":
        lut.SetColorSpaceToDiverging()
    elif color_space == "HSV":
        lut.SetColorSpaceToHSV()
    elif color_space == "Lab":
        lut.SetColorSpaceToLab()
    elif color_space == "RGB":
        lut.SetColorSpaceToRGB()
    elif color_space == "CIELAB":
        lut.SetColorSpaceToLabCIEDE2000()

    if "NanColor" in preset:
        lut.SetNanColor(preset["NanColor"])

    # Always RGB points
    lut.RemoveAllPoints()
    for arr in np.split(srgb, len(srgb) / 4):
        lut.AddRGBPoint(arr[0], arr[1], arr[2], arr[3])

    return lut


def set_preset(lut: vtkLookupTable, preset_name: str, n_colors=255):
    colors = get_preset(preset_name)
    min, max = colors.GetRange()
    delta = max - min
    lut.SetNumberOfTableValues(n_colors)
    for i in range(n_colors):
        x = min + (delta * i / n_colors)
        rgb = colors.GetColor(x)
        lut.SetTableValue(i, *rgb)
    lut.Build()



import json
import traceback
from pathlib import Path

import numpy as np

from pan3d import catalogs as pan3d_catalogs
from pan3d.ui.collapsible import CollapsableSection
from pan3d.ui.css import base, preview
from pan3d.utils.constants import SLICE_VARS, XYZ
from pan3d.utils.convert import update_camera
from pan3d.widgets.color_by import ColorBy
from pan3d.xarray.algorithm import vtkXArrayRectilinearSource
from trame.app import TrameApp, asynchronous
from trame.decorators import change
from trame.widgets import html
from trame.widgets import vuetify3 as v3


class Explorer(TrameApp):
    def __init__(self, xarray=None, source=None, server=None, local_rendering=None):
        """
        Parameters:
            server (str/server): Trame server name or instance.
            local_rendering (str): If provided (wasm, vtkjs) local rendering will be used

        CLI options:
            - `--import-state`: Pass a string with this argument to specify a startup configuration.
                              This value must be a local path to a JSON file which adheres to the
                              schema specified in the [Configuration Files documentation](../api/configuration.md).
                              A dataset specified in this configuration will override any value passed to `--xarray-*`
            - `--xarray-file`: Provide path to xarray file
            - `--xarray-url`: Provide URL to xarray dataset
            - `--wasm`: Use WASM for local rendering
            - `--vtkjs`: Use vtk.js for local rendering
        """
        super().__init__(server, client_type="vue3")

        parser = self.server.cli
        explorer = parser.add_argument_group("Explorer Properties")
        explorer.add_argument(
            "--import-state",
            help="Pass a string with this argument to specify a startup configuration. This value must be a local path to a JSON file which adheres to the schema specified in the [Configuration Files documentation](../api/configuration.md). A dataset specified in this configuration will override any value passed to `--xarray-*`",
        )
        explorer.add_argument(
            "--xarray-file",
            help="Provide path to xarray file",
        )
        explorer.add_argument(
            "--xarray-url",
            help="Provide URL to xarray dataset",
        )

        rendering = parser.add_argument_group("Rendering Properties")
        # Local rendering setup
        rendering.add_argument(
            "--wasm",
            help="Use WASM for local rendering",
            action="store_true",
        )
        rendering.add_argument(
            "--vtkjs",
            help="Use vtk.js for local rendering",
            action="store_true",
        )

        # CLI
        args, _ = self.server.cli.parse_known_args()
        # Local rendering
        self.local_rendering = local_rendering
        if args.wasm:
            self.local_rendering = "wasm"
        if args.vtkjs:
            self.local_rendering = "vtkjs"

        self.state.nan_colors = [
            [0, 0, 0, 1],
            [0.99, 0.99, 0.99, 1],
            [0.6, 0.6, 0.6, 1],
            [1, 0, 0, 1],
            [0, 1, 0, 1],
            [0, 0, 1, 1],
            [0.9, 0.9, 0.9, 0],
        ]
        self.state.nan_color = 2

        self.ui = None

        # Initialize source
        self.xarray = None
        self.source = None

        if source is not None:
            self.source = source
            self.xarray = source.input
        elif xarray is not None:
            self.source = vtkXArrayRectilinearSource(input=xarray)
            self.xarray = xarray

        # Process CLI
        self.ctrl.on_server_ready.add(self._process_cli)

    # -------------------------------------------------------------------------
    # Trame API
    # -------------------------------------------------------------------------

    def _process_cli(self, **_):
        args, _ = self.server.cli.parse_known_args()

        # import state
        if args.import_state:
            self._import_file_from_path(args.import_state)

        # load xarray (file)
        elif args.xarray_file:
            self.state.import_pending = True
            with self.state:
                self.load_dataset("file", args.xarray_file)
                self.state.data_origin_id = str(Path(args.xarray_file).resolve())
            self.state.import_pending = False

        # load xarray (url)
        elif args.xarray_url:
            self.state.import_pending = True
            with self.state:
                self.load_dataset("url", args.xarray_url)
                self.state.data_origin_id = args.xarray_url
            self.state.import_pending = False

        # Load given XArray
        elif self.xarray is not None:
            self.state.show_data_information = True
            self.ctrl.xr_update_info(self.source.input, self.source.available_arrays)
            self.ctx.rendering.update_from_source(self.source)

    def start(self, **kwargs):
        """Initialize the UI and start the server for XArray Viewer."""
        self.ui.server.start(**kwargs)

    @property
    async def ready(self):
        """Start and wait for the XArray Viewer corroutine to be ready."""
        await self.ui.ready

    @change("data_origin_order")
    def _on_order_change(self, **_):
        if self.state.import_pending:
            return

        self.state.load_button_text = "Load"
        self.state.can_load = True

    @change("data_origin_source")
    def _on_data_origin_source(self, data_origin_source, **kwargs):
        if self.state.import_pending:
            return

        self.state.data_origin_id = ""
        results, *_ = pan3d_catalogs.search(data_origin_source)
        self.state.data_origin_ids = [v["name"] for v in results]
        self.state.data_origin_id_to_desc = {
            v["name"]: v["description"] for v in results
        }

    @change("data_origin_id")
    def _on_data_origin_id(self, data_origin_id, data_origin_source, **kwargs):
        if self.state.import_pending:
            return

        self.state.load_button_text = "Load"
        self.state.can_load = True

        if data_origin_source == "file":
            self.state.data_origin_id_error = not Path(data_origin_id).exists()
        elif self.state.data_origin_id_error:
            self.state.data_origin_id_error = False

    # -----------------------------------------------------
    # UI Components
    # -----------------------------------------------------
    @change("color_by", "color_preset", "color_min", "color_max", "nan_color")
    def _on_color_properties_change(self, **_):
        if self.mapper:
            self.ctx.rendering.color_by.configure_mapper(self.mapper)
            self.ctx.scalar_bar.preset = self.state.color_preset
            self.ctx.scalar_bar.set_color_range(
                self.state.color_min, self.state.color_max
            )
            self.ctrl.view_update()

    @change("slice_t", *[var.format(axis) for axis in XYZ for var in SLICE_VARS])
    def on_change(self, slice_t, **_):
        source = self.source
        if source is None:
            return

        if self.state.import_pending:
            return

        slices = {source.t: slice_t}
        for axis in XYZ:
            axis_name = getattr(source, axis)
            if axis_name is None:
                continue

            if self.state[f"slice_{axis}_type"] == "range":
                if self.state[f"slice_{axis}_range"] is None:
                    continue
                slices[axis_name] = [
                    *self.state[f"slice_{axis}_range"],
                    int(self.state[f"slice_{axis}_step"]),
                ]
                slices[axis_name][1] += 1  # end is exclusive
            else:
                slices[axis_name] = self.state[f"slice_{axis}_cut"]

        source.slices = slices
        ds = source()
        self.state.dataset_bounds = ds.bounds

        self.ctrl.view_reset_clipping_range()
        self.ctrl.view_update()

    @change("slice_t")
    def _on_slice_t(self, slice_t, **_):
        if self.state.import_pending:
            return

        self.source.t_index = slice_t
        self.ctrl.view_update()

    # -----------------------------------------------------
    # Triggers
    # -----------------------------------------------------

    def import_file_upload(self, files):
        self.import_state(json.loads(files[0].get("content")))

    def _import_file_from_path(self, file_path):
        if file_path is None:
            return

        file_path = Path(file_path)
        if file_path.exists():
            self.import_state(json.loads(file_path.read_text("utf-8")))

    def load_dataset(self, source, id, order="C", config=None):
        self.state.data_origin_source = source
        self.state.data_origin_id = id
        self.state.load_button_text = "Loaded"
        self.state.can_load = False
        self.state.show_data_information = True

        if config is None:
            config = {
                "arrays": [],
                "slices": {},
            }

        try:
            self.source.load(
                {
                    "data_origin": {
                        "source": source,
                        "id": id,
                        "order": order,
                    },
                    "dataset_config": config,
                }
            )

            # Extract UI
            self.ctrl.xr_update_info(self.source.input, self.source.available_arrays)
            self.ctx.rendering.update_from_source(self.source)

            # no error
            self.state.data_origin_error = False
        except Exception as e:
            self.state.data_origin_error = (
                f"Error occurred while trying to load data. {e}"
            )
            self.state.data_origin_id_error = True
            self.state.load_button_text = "Load"
            self.state.can_load = True
            self.state.show_data_information = False

            print(traceback.format_exc())

    def update_rendering(self, reset_camera=False):
        raise NotImplementedError(
            """
            This method is supposed to be implemented by the Explorer.
            Please override this in the necessary Explorer class if you see this message
            """
        )

    # -----------------------------------------------------
    # Public API
    # -----------------------------------------------------

    def export_state(self, **kwargs):
        """Return a json dump of the reader and viewer state"""
        camera = self.renderer.active_camera
        state_to_export = {
            **self.source.state,
            "rendering": {
                "view_3d": self.state.view_3d,
                "color_by": self.state.color_by,
                "color_preset": self.state.color_preset,
                "color_min": self.state.color_min,
                "color_max": self.state.color_max,
                "scale_x": self.state.scale_x,
                "scale_y": self.state.scale_y,
                "scale_z": self.state.scale_z,
            },
            "camera": {
                "position": camera.position,
                "view_up": camera.view_up,
                "focal_point": camera.focal_point,
                "parallel_projection": camera.parallel_projection,
                "parallel_scale": camera.parallel_scale,
            },
            **kwargs,
        }
        return json.dumps(state_to_export, indent=2)

    def import_state(self, data_state):
        """
        Read the current state to load the data and visualization setup if any.

        Parameters:
            - data_state (dict): reader (+viewer) state to reset to
        """
        self.state.import_pending = True
        try:
            data_origin = data_state.get("data_origin")
            source = data_origin.get("source")
            id = data_origin.get("id")
            order = data_origin.get("order", "C")
            config = data_state.get("dataset_config")
            rendering_state = data_state.get("rendering", {})
            camera_state = data_state.get("camera", {})

            # load data and initial rendering setup
            with self.state:
                self.load_dataset(source, id, order, config)
                self.state.update(rendering_state)

            # override computed color range using state values
            with self.state:
                self.state.update(rendering_state)

            # update camera and render
            update_camera(self.renderer.active_camera, camera_state)
            self.update_rendering()
        finally:
            self.state.import_pending = False

    async def _save_dataset(self, file_path):
        output_path = Path(file_path).resolve()
        self.source.input.to_netcdf(output_path)

    def save_dataset(self, file_path):
        """
        Write XArray data into a file using a background task.
        So when used programmatically, make sure you await the returned task.

        Parameters:
            - file_path (str): path to use for writing the file

        Returns:
            writing task
        """
        self.state.show_save_dialog = False
        return asynchronous.create_task(self._save_dataset(file_path))


class SummaryToolbar(v3.VCard):
    def __init__(
        self,
        t_labels="t_labels",
        slice_t="slice_t",
        slice_t_max="slice_t_max",
        color_by="color_by",
        data_arrays="data_arrays",
        max_time_width="max_time_width",
        max_time_index_width="max_time_index_width",
        **kwargs,
    ):
        super().__init__(
            classes="summary-toolbar",
            rounded="pill",
            **kwargs,
        )

        # Activate CSS
        self.server.enable_module(base)
        self.server.enable_module(preview)

        with self:
            with v3.VToolbar(
                classes="pl-2",
                height=50,
                elevation=1,
                style="background: none;",
            ):
                v3.VIcon("mdi-clock-outline")
                html.Pre(
                    f"{{{{ {t_labels}[slice_t] }}}}",
                    classes="mx-2 text-left",
                    style=(f"`min-width: ${{ {max_time_width} }}rem;`",),
                )
                v3.VSlider(
                    prepend_inner_icon="mdi-clock-outline",
                    v_model=(slice_t, 0),
                    min=0,
                    max=(slice_t_max, 0),
                    step=1,
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    classes="mx-2",
                )
                html.Div(
                    f"{{{{ {slice_t} + 1 }}}}/{{{{ {slice_t_max} + 1 }}}}",
                    classes="mx-2 text-right",
                    style=(f"`min-width: ${{ {max_time_index_width} }}rem;`",),
                )
                v3.VSelect(
                    placeholder="Color By",
                    prepend_inner_icon="mdi-format-color-fill",
                    v_model=(color_by, None),
                    items=(data_arrays, []),
                    clearable=True,
                    hide_details=True,
                    density="compact",
                    flat=True,
                    variant="solo",
                    max_width=200,
                )


class DataOrigin(CollapsableSection):
    def __init__(self, load_dataset):
        super().__init__("Data origin", "show_data_origin", True)

        self.state.load_button_text = "Load"
        self.state.can_load = True
        self.state.data_origin_id_to_desc = {}

        with self.content:
            v3.VSelect(
                label="Source",
                v_model=("data_origin_source", "xarray"),
                items=(
                    "data_origin_sources",
                    pan3d_catalogs.list_availables(),
                ),
                hide_details=True,
                density="compact",
                flat=True,
                variant="solo",
            )
            v3.VDivider()
            v3.VTextField(
                placeholder="Location",
                v_if="['file', 'url'].includes(data_origin_source)",
                v_model=("data_origin_id", ""),
                hide_details=True,
                density="compact",
                flat=True,
                variant="solo",
                append_inner_icon=(
                    "data_origin_id_error ? 'mdi-file-document-alert-outline' : undefined",
                ),
                error=("data_origin_id_error", False),
            )

            with v3.VTooltip(
                v_else=True,
                text=("`${ data_origin_id_to_desc[data_origin_id] }`",),
            ):
                with html.Template(v_slot_activator="{ props }"):
                    v3.VSelect(
                        v_bind="props",
                        label="Name",
                        v_model="data_origin_id",
                        items=("data_origin_ids", []),
                        hide_details=True,
                        density="compact",
                        flat=True,
                        variant="solo",
                    )
            v3.VDivider()
            v3.VBtn(
                "{{ load_button_text }}",
                block=True,
                classes="text-none",
                flat=True,
                density="compact",
                rounded=0,
                disabled=("!data_origin_id?.length || !can_load",),
                color=("can_load ? 'primary': undefined",),
                click=(
                    load_dataset,
                    "[data_origin_source, data_origin_id, data_origin_order]",
                ),
            )


class DataInformation(CollapsableSection):
    def __init__(self, xarray_info="xarray_info"):
        super().__init__("Data information", "show_data_information")

        self._var_name = xarray_info
        self.state.setdefault(xarray_info, [])

        with self.content:
            with v3.VTable(density="compact", hover=True):
                with html.Tbody():
                    with html.Template(v_for=f"item, i in {xarray_info}", key="i"):
                        with v3.VTooltip():
                            with html.Template(v_slot_activator="{ props }"):
                                with html.Tr(v_bind="props", classes="pointer"):
                                    with html.Td(
                                        classes="d-flex align-center text-no-wrap"
                                    ):
                                        v3.VIcon(
                                            "{{ item.icon }}",
                                            size="sm",
                                            classes="mr-2",
                                        )
                                        html.Div("{{ item.name }}")
                                    html.Td(
                                        "{{ item.length }}",
                                        classes="text-right",
                                    )

                            with v3.VTable(
                                density="compact",
                                theme="dark",
                                classes="no-bg ma-0 pa-0",
                            ):
                                with html.Tbody():
                                    with html.Tr(
                                        v_for="attr, j in item.attrs",
                                        key="j",
                                    ):
                                        html.Td(
                                            "{{ attr.key }}",
                                        )
                                        html.Td(
                                            "{{ attr.value }}",
                                        )

    def update_information(self, xr, available_arrays=None):
        xarray_info = []
        coords = set(xr.coords.keys())
        data = set(available_arrays or [])
        for name in xr.variables:
            icon = "mdi-variable"
            order = 3
            length = f"({','.join(xr[name].dims)})"
            attrs = []
            if name in coords:
                icon = "mdi-ruler"
                order = 1
                length = xr[name].size
                shape = xr[name].shape
                if length > 1 and len(shape) == 1:
                    attrs.append(
                        {
                            "key": "range",
                            "value": f"[{xr[name].values[0]}, {xr[name].values[-1]}]",
                        }
                    )
            if name in data:
                icon = "mdi-database"
                order = 2
            xarray_info.append(
                {
                    "order": order,
                    "icon": icon,
                    "name": name,
                    "length": length,
                    "type": str(xr[name].dtype),
                    "attrs": attrs
                    + [
                        {"key": "type", "value": str(xr[name].dtype)},
                    ]
                    + [
                        {"key": str(k), "value": str(v)}
                        for k, v in xr[name].attrs.items()
                    ],
                }
            )
        xarray_info.sort(key=lambda item: item["order"])

        # Update UI
        self.state[self._var_name] = xarray_info


class ControlPanel(v3.VCard):
    def __init__(
        self,
        enable_data_selection,
        toggle,
        load_dataset,
        import_file_upload,
        export_file_download,
        xr_update_info="xr_update_info",
        panel_label="XArray Viewer",
        **kwargs,
    ):
        super().__init__(
            classes="controller",
            rounded=(f"{toggle} || 'circle'",),
            **kwargs,
        )

        # state initialization
        self.state.import_pending = False

        # extract trigger name
        download_export = self.ctrl.trigger_name(export_file_download)

        with self:
            with v3.VCardTitle(
                classes=(
                    f"`d-flex pa-1 position-fixed bg-white ${{ {toggle} ? 'controller-content rounded-t border-b-thin':'rounded-circle'}}`",
                ),
                style="z-index: 1;",
            ):
                v3.VProgressLinear(
                    v_if=toggle,
                    indeterminate=("trame__busy",),
                    bg_color="rgba(0,0,0,0)",
                    absolute=True,
                    color="primary",
                    location="bottom",
                    height=2,
                )
                v3.VProgressCircular(
                    v_else=True,
                    bg_color="rgba(0,0,0,0)",
                    indeterminate=("trame__busy",),
                    style="position: absolute; top: 0; left: 0; width: 100%; height: 100%;",
                    color="primary",
                    width=3,
                )
                v3.VBtn(
                    icon="mdi-close",
                    v_if=toggle,
                    click=f"{toggle} = !{toggle}",
                    flat=True,
                    size="sm",
                )
                v3.VBtn(
                    icon="mdi-menu",
                    v_else=True,
                    click=f"{toggle} = !{toggle}",
                    flat=True,
                    size="sm",
                )
                if self.server.hot_reload:
                    v3.VBtn(
                        v_show=toggle,
                        icon="mdi-refresh",
                        flat=True,
                        size="sm",
                        click=self.ctrl.on_server_reload,
                    )
                v3.VSpacer()
                html.Div(
                    panel_label,
                    v_show=toggle,
                    classes="text-h6 px-2",
                )
                v3.VSpacer()

                with v3.VMenu(v_if=toggle, density="compact"):
                    with html.Template(v_slot_activator="{props}"):
                        v3.VBtn(
                            v_bind="props",
                            icon="mdi-file-arrow-left-right-outline",
                            flat=True,
                            size="sm",
                            classes="mx-1",
                        )
                    with v3.VList(density="compact"):
                        if enable_data_selection:
                            with v3.VListItem(
                                title="Export state file",
                                disabled=("can_load",),
                                click=f"utils.download('xarray-state.json', trigger('{download_export}'), 'text/plain')",
                            ):
                                with html.Template(v_slot_prepend=True):
                                    v3.VIcon(
                                        "mdi-cloud-download-outline", classes="mr-n5"
                                    )

                        with v3.VListItem(
                            title="Import state file",
                            click="trame.utils.get('document').querySelector('#fileImport').click()",
                        ):
                            html.Input(
                                id="fileImport",
                                hidden=True,
                                type="file",
                                change=(
                                    import_file_upload,
                                    "[$event.target.files]",
                                ),
                                __events=["change"],
                            )
                            with html.Template(v_slot_prepend=True):
                                v3.VIcon("mdi-cloud-upload-outline", classes="mr-n5")
                        v3.VDivider()
                        with v3.VListItem(
                            title="Save dataset to disk",
                            disabled=("can_load",),
                            click="show_save_dialog = true",
                        ):
                            with html.Template(v_slot_prepend=True):
                                v3.VIcon("mdi-file-download-outline", classes="mr-n5")

            with v3.VCardText(
                v_show=(toggle, True),
                classes="controller-content py-1 mt-10",
            ) as ui_content:
                self.ui_content = ui_content
                if enable_data_selection:
                    DataOrigin(load_dataset)
                self.ctrl[xr_update_info] = DataInformation().update_information


class RenderingSettingsBasic(CollapsableSection):
    def __init__(self, source=None, update_rendering=None, **kwargs):
        super().__init__(self, "Rendering", "show_rendering", **kwargs)
        self.source = source

        with self.content:
            v3.VSelect(
                placeholder="Data arrays",
                prepend_inner_icon="mdi-database",
                hide_selected=True,
                v_model=("data_arrays", []),
                items=("data_arrays_available", []),
                multiple=True,
                hide_details=True,
                density="compact",
                chips=True,
                closable_chips=True,
                flat=True,
                variant="solo",
            )
            v3.VDivider()
            self.color_by = ColorBy(
                color_by_name="color_by",
                preset_name="color_preset",
                color_min_name="color_min",
                color_max_name="color_max",
                nan_color_name="nan_color",
                reset_color_range=self.reset_color_range,
            )

    def reset_color_range(self):
        """Reset the color range to the min and max values of the selected data array."""
        color_by = self.color_by.color_by
        ds = self.source()
        array = (
            ds.point_data[color_by]
            if color_by in ds.point_data.keys()
            else ds.cell_data[color_by]
            if color_by in ds.cell_data.keys()
            else None
        )
        if array is not None:
            self.color_by.color_min = float(np.min(array))
            self.color_by.color_max = float(np.max(array))
        else:
            self.color_by.color_min = 0.0
            self.color_by.color_max = 1.0

        self.ctrl.view_update()

    @change("data_arrays")
    def _on_array_selection(self, data_arrays, **_):
        # if self.state.import_pending:
        #    return
        self.state.dirty_data = True
        if self.source is not None:
            self.source.arrays = data_arrays

        if self.source is None or self.source.input is None:
            self.color_by.data_arrays = []
        else:
            self.color_by.set_data_arrays_from_vtk(self.source())

    def update_from_source(self, source=None):
        raise NotImplementedError(
            """
            This method needs to be implemented in the specialization of this class.
            Please override it in the necessary class representing the rendering settings for the Explorer.
            """
        )
